/*
 *   filesystem test
 *
 *   TODO: Use openat?
 */
#include <locale.h>
#include <libintl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <pwd.h>
#include <grp.h>
#include <assert.h>
#include <mntent.h>
#include <errno.h>
#include <ctype.h>
#include <fnmatch.h>
#include <sys/wait.h>
#include <sectool.h>
#include <fcntl.h>
#include <rpmdb.h>
#include <rpmts.h>
#include <rpmlib.h>
#include <rpmfileutil.h>
#include <rpmmacro.h>

#ifdef TEST_SELINUX
#include <selinux/selinux.h>
#endif

#ifdef DEBUG
#include <time.h>
static time_t t_start = 0, t_end = 0;
#endif

static rpmts RPMts;

static char *myname = NULL;
static char *dirs_include = NULL;
static char *dirs_exclude = NULL;
static char **dirs_exclude_arr = NULL;
static size_t dirs_exclude_arr_size = 0;
static char *rpmcheck_enabled = NULL;
static char *selinuxcheck_enabled = NULL;
static char *fstypes = NULL;
static char **fsstr_arr = NULL;
static size_t fsstr_arr_size = 0;
static char *symwhite = NULL;
static char **symwhite_arr = NULL;
static size_t symwhite_arr_size = 0;
static char **symwhite_tarr = NULL;
static size_t symwhite_tarr_size = 0;
static int do_selinux_check = 0;

struct envinit_t vars[]={
  {"NAME"    , &myname},
  {"INCLUDE" , &dirs_include},
  {"EXCLUDE" , &dirs_exclude},
  {"RPMCHECK", &rpmcheck_enabled},
  {"FSTYPES" , &fstypes},
  {"SELINUXCHECK", &selinuxcheck_enabled},
  {"SYMLINK_WHITELIST", &symwhite},
  { NULL, NULL}
};

/*
 *  Error IDs
 */
#define ID_DIRWORLDW        1
#define ID_FILEGROUPWX      2
#define ID_FILEWORLDW       3
#define ID_GROUPNOTFOUND    4
#define ID_NOPKG            5
#define ID_SUIDSCRIPT       6
#define ID_USERNOTFOUND     7
#define ID_WRONGSYMLINK     8
#define ID_FILEWORLDWX      9
#define ID_SE_MISLABELED    10
//#define ID_SE_DEVMISLABELED 11
#define ID_CANTCHECK        12

#define MAX_DIR_STACK_SIZE  16384

#ifdef DEBUG
static size_t checked_objects = 0;
#endif

#define INIT_DEVID_SIZE 64
#define DEVID_GROW_COEF 1.2

static dev_t  *DEVid = NULL;
static size_t  DEVid_count = 0;
static size_t  DEVid_size  = 0;

#define MTAB_PATH "/etc/mtab"

char *ltrim (char *str) {
  while (*str == ' ' || *str == '\t')
    ++str;
  return str;
}

int devidcmp (const void *a, const void *b) {
  return ((*(const dev_t *)a)>=(*(const dev_t *)b))?((*(const dev_t *)a)==(*(const dev_t *)b)?0:1):-1;
}

int matchfs (const char *fs) {
  register size_t i;

  for (i = 0; i < fsstr_arr_size; ++i) {
    if (!strcmp (fs, fsstr_arr[i]))
      return 1;
  }
  
  return 0;
}

void init_fsstr (void) {
  char *fstype;
  
  if (fstypes != NULL) {
    fstype = strtok (fstypes, " ");

    do {
      fsstr_arr = xrealloc (fsstr_arr, sizeof (char *) * (fsstr_arr_size + 1));
      fsstr_arr[fsstr_arr_size++] = strdup (fstype);
    } while ((fstype = strtok (NULL, " ")) != NULL);
  }
  
  return;
}

void free_fsstr (void) {
  for (; fsstr_arr_size > 0; --fsstr_arr_size) {
    free (fsstr_arr[fsstr_arr_size - 1]);
  }

  free (fsstr_arr);
  fsstr_arr = NULL;

  return;
}


int matchsymwhite (const char *symlink, const char *target) {
  register size_t i;
  
  for (i = 0; i < symwhite_arr_size; ++i) {
    if (fnmatch ( symwhite_arr[i], symlink, 0) == 0 &&
	fnmatch (symwhite_tarr[i],  target, 0) == 0) {
      return 1;
    }
  }
  
  return 0;
}

void init_symwhite (void) {
  char *symlink, *target;
  
  if (symwhite != NULL) {
    symlink = strtok (symwhite, "\n");
    
    do {
      symwhite_tarr = xrealloc (symwhite_tarr, sizeof (char *) * (symwhite_tarr_size + 1));
      symwhite_arr  = xrealloc (symwhite_arr,  sizeof (char *) * (symwhite_arr_size + 1));

      target = strstr (symlink, " -> ");
      if (target != NULL) {
	target[0] = '\0';
	target = target+4;

	symwhite_arr[symwhite_arr_size++]  = strdup (symlink);
	target = ltrim (target);

	if (target[0] != '\0') {
	  symwhite_tarr[symwhite_tarr_size++] = strdup (target);
	} else {
	  test_exit (E_FAIL, _("Syntax error in SYMLINK_WHITELIST, expecting target after \" -> \".\n"));
	}
	
      } else {
	symwhite_arr[symwhite_arr_size++]  = strdup (symlink);
	symwhite_tarr[symwhite_tarr_size++] = strdup ("*");
      }

      //fprintf (stderr, "S: \"%s\" -> \"%s\"\n", symwhite_arr[symwhite_arr_size - 1], symwhite_tarr[symwhite_tarr_size-1]);
      
    } while ((symlink = strtok (NULL, "\n")) != NULL);
  }
  
  return;
}

void free_symwhite (void) {
  for (; symwhite_arr_size > 0; --symwhite_arr_size) {
    free (symwhite_arr[symwhite_arr_size - 1]);
    free (symwhite_tarr[symwhite_arr_size - 1]);
  }

  free (symwhite_tarr);
  free (symwhite_arr);
  symwhite_arr = NULL;
  symwhite_tarr = NULL;

  return;
}

void init_devs (const char *mtab) {
  struct mntent *mnt;
  FILE *fp;
  struct stat st;

  init_fsstr ();

  fp = setmntent (mtab, "r");
  if (fp == NULL) {
#ifdef DEBUG
    fprintf (stderr, "setmntent: %s: %s\n", mtab, strerror (errno));
#endif
    test_exit (E_FAIL, _("setmntent(%s): %s\n"), mtab, strerror (errno));
  }
  
  DEVid = xmalloc (sizeof (dev_t) * INIT_DEVID_SIZE);
  DEVid_size = INIT_DEVID_SIZE;
  DEVid_count = 0;
  
  while ((mnt = getmntent (fp)) != NULL) {
    if (matchfs (mnt->mnt_type)) {      
      if (lstat (mnt->mnt_dir, &st) != 0) {
#ifdef DEBUG
        fprintf (stderr, "lstat: %s: %s\n", mnt->mnt_dir, strerror (errno));
#endif
	test_exit (E_FAIL, _("lstat(%s): %s\n"), mnt->mnt_dir, strerror (errno));
      }

#ifdef DEBUG
      fprintf (stderr, "m: %s, t: %s, i: %llu\n", mnt->mnt_dir, mnt->mnt_type, st.st_dev);
#endif

      DEVid[DEVid_count] = st.st_dev;
      ++DEVid_count;

      if (DEVid_count == DEVid_size) {
	DEVid_size *= DEVID_GROW_COEF;
	DEVid = xrealloc (DEVid, sizeof (unsigned int) * DEVid_size);
      }
    }
  }

  if (DEVid_count == 0) {
    if (fstypes != NULL) {
      test_exit (E_FAIL, _("No local devices found. This is probably due to a misconfiguration. Check the FSTYPES variable in dsc file.\n"));
    } else {
      test_exit (E_FAIL, _("Please set the FSTYPES variable in dsc file.\n"));
    }
  }

  qsort (DEVid, DEVid_count, sizeof (dev_t), &devidcmp);

#ifdef DEBUG
  { register size_t i;
    for (i = 0; i < DEVid_count; ++i)
      fprintf (stderr, "DEV: %llu\n", DEVid[i]); 
  }
#endif

  fclose (fp);
  free_fsstr ();

  return;
}

#ifdef DEBUG
void print_env (struct envinit_t *e) {
  size_t i;

  for (i = 0; e[i].env_var != NULL; ++i)
    fprintf (stderr, "%s = \"%s\"\n", e[i].env_var, *(e[i].varptr));
  
  return;
}
#endif

const char *mode2str (mode_t m) {
  switch (m & S_IFMT) {
  case S_IFBLK:  return _("block device");
  case S_IFCHR:  return _("character device");
  case S_IFDIR:  return _("directory");
  case S_IFIFO:  return _("FIFO/pipe");
  case S_IFLNK:  return _("symlink");
  case S_IFREG:  return _("regular file");
  case S_IFSOCK: return _("socket");
  default:       return _("unknown");
  }
}

typedef int(*strcmp_func_t)(const char*, const char*);

/*
 * Return 1 if param what matches any of strings of NULL-terminated
 * against, with respect to comp_func )function used for comparison)
 */
int string_match_any(const char* what, const char** against, strcmp_func_t cmp)
{
	for (; *against; ++against)
		if (cmp(what, *against) == 0)
			return 1;
	return 0;
}

int option_enabled(const char* optval)
{
	static const char* enabled_vals[]  = {"1", "yes", "true", NULL};
	// static const char* disabled_vals = {"0", "no",  "false", NULL};
	return string_match_any(optval, enabled_vals, strcasecmp);
}

/*
 *  Main
 */
 
int isexcluded (const char *path) {
  register size_t i;
  
  if (dirs_exclude_arr != NULL) {
    for (i = 0; i < dirs_exclude_arr_size; ++i) {
      if (!fnmatch (dirs_exclude_arr[i], path, FNM_PATHNAME)) {
	return 1;
      }
    }
  }
  return 0;
}

int islocal (dev_t devid) {
  register size_t l,r,d;

  assert (DEVid_count > 0);

  l = 0;
  r = DEVid_count - 1;
  d = DEVid_count;

  while (l != r) {
    if (devid < DEVid[l+d/2]) {
      r -= d - d/2;
    } else {
      l += d/2;
    }    
    d = r - l + 1;
  }
  
  return (DEVid[l] == devid);
}

#ifdef TEST_SELINUX

void check_selinux (const char *fpath, const struct stat *s)
{
	if ((s->st_mode & S_IFMT) == S_IFLNK)
		return;

	if (selinux_file_context_verify(fpath, s->st_mode) == 0) {
		security_context_t filecontext = NULL, pathcontext = NULL;
		int pres;

		getfilecon(fpath, &filecontext);
		pres = matchpathcon(fpath, 1, &pathcontext);

		if (pres == 0) {
			report(WARNING, ID_SE_MISLABELED, _("Mislabeled %s '%s' found. Labeled as '%s', should be '%s'.\n"),
				mode2str(s->st_mode), fpath, filecontext, pathcontext);
			report(HINT,    ID_SE_MISLABELED, _("File is not labeled as defined in configuration. See man restorecon.\n"));
		} else { /* TODO */ }

		freecon(filecontext);
		freecon(pathcontext);
	}
}

#endif // TEST_SELINUX

void check_object (const char *path, const struct stat *s) {
  struct passwd *pw;
  struct group  *gr;

#ifdef DEBUG
  ++checked_objects;
#endif
  
  pw = getpwuid (s->st_uid);
  if (pw == NULL) {
    report (ERROR, ID_USERNOTFOUND,
	    _("User ID of %s \"%s\" doesn't exist in user database.\n"), mode2str(s->st_mode), path);
  }

  gr = getgrgid (s->st_gid);
  if (gr == NULL) {
    report (ERROR, ID_GROUPNOTFOUND,
	    _("Group ID of %s \"%s\" doesn't exist in group database.\n"), mode2str(s->st_mode), path);
  }

  switch (s->st_mode & S_IFMT)
    {
    case S_IFREG: /* regular file */
      
      /* world writable */
      if (s->st_mode & S_IWOTH) {
	report (WARNING, ID_FILEWORLDW, _("File \"%s\" is world writable.\n"), path);
      }
      
      /* executable and world/group writable */
      if ((s->st_mode&(S_IXUSR|S_IXGRP|S_IXOTH))) {
	if (s->st_mode&(S_IWOTH)) {
	  report (ERROR, ID_FILEWORLDWX, _("File \"%s\" is executable and world writable.\n"), path);
	} else if (s->st_mode&(S_IWGRP)) {
	  report (WARNING, ID_FILEGROUPWX, _("File \"%s\" is executable and group writable.\n"), path);
	}	
      }

      /* SUID script */
      if (((s->st_mode&(S_IXUSR|S_IXGRP|S_IXOTH)) && (s->st_mode&(S_ISUID))) ||
	  ((s->st_mode&(S_IXGRP)) && (s->st_mode&(S_ISUID)))) {

	{ FILE *fp;
	  char mb[2];
	  fp = fopen (path, "r");
	  if (fp != NULL) {
	    if (fread (mb, 1, 2, fp) == 2) {
	      if (mb[0] == '#' && mb[1] == '!') {
		report (ERROR, ID_SUIDSCRIPT, _("File \"%s\" is SUID script.\n"), path);
	      }
	    }
	    /* FIXME: verbose? */
	    fclose (fp);
	  }}
      }

      /* rpmquery */
      if ((s->st_mode&(S_IXUSR|S_IXGRP|S_IXOTH)) && strcmp (rpmcheck_enabled, "0")) {
	char *cpath = strdup (path);
	(void) rpmCleanPath (cpath);
	
	if (rpmtsInitIterator(RPMts, RPMTAG_BASENAMES, path, 0)   == NULL &&
	    rpmtsInitIterator(RPMts, RPMTAG_PROVIDENAME, path, 0) == NULL) {
	  report (ERROR, ID_NOPKG, _("The executable file \"%s\" doesn't belong to any package.\n"), path);
	}

	free (cpath);
      }
      
      break;
    case S_IFDIR: /* directory */
      
      /* world writable directory without sticky bit */
      if ((s->st_mode&(S_IWOTH)) &&
	  !(s->st_mode&(S_ISVTX))) {
	report (WARNING, ID_DIRWORLDW, _("Directory \"%s\" is world writable but the sticky bit is not set.\n"), path);
      }
      
      break;
    case S_IFLNK: /* symbolic link */
      
      /* non-existent target */
      { struct stat sl;
	
	if (stat (path, &sl) != 0) {
	  char target[PATH_MAX+1];
	  size_t tsize;
	  
	  tsize = readlink (path, target, PATH_MAX);
	  target[tsize] = '\0';
	  
	  if (!matchsymwhite (path, target)) {
	    report (WARNING, ID_WRONGSYMLINK, _("Symbolic link \"%s\" points to a non-existent file \"%s\".\n"), path, target); 
	  }
	}
      }
      
      break;
    }

#ifdef TEST_SELINUX
    if (do_selinux_check)
      check_selinux(path, s);
#endif

  return;
}

void walkdir (const char *path, void (*check)(const char *, const struct stat *)) {
  DIR *dp;
  struct dirent *item;
  char item_path[PATH_MAX];
  const char *mypath = path;
  //char **dirs_stack = NULL;
  //register size_t dirs_stack_size = 0;
  struct stat st;
  
  assert (strlen (path) > 0);

  /* Very nasty! */
  if (path[0] == '/' && path[1] == '\0') {
    mypath = "";
  }
  
  dp = opendir (path);
  if (dp != NULL) {
    //item_path = xmalloc (sizeof (char) * PATH_MAX);
    
    while ((item = readdir (dp)) != NULL) {
      if (strcmp (item->d_name,  ".") != 0 &&
	  strcmp (item->d_name, "..") != 0) {
	
	sprintf (item_path, "%s/%s", mypath, item->d_name);
	
	if (lstat (item_path, &st) == 0) {
	  check (item_path, &st);
	} else {
#ifdef DEBUG
	  fprintf (stderr, "lstat: %s: %s\n", item_path, strerror (errno));
#endif
	  report (WARNING, ID_CANTCHECK, _("Can't get file status for path \"%s\".\n"), item_path);
	  //test_exit (E_FAIL, _("lstat(%s): %s\n"), item_path, strerror (errno));
	}
	
	/*
	  if (S_ISDIR(st.st_mode) &&
	  !isexcluded (item_path) && islocal (st.st_dev) &&
	  dirs_stack_size < MAX_DIR_STACK_SIZE) {
	  ++dirs_stack_size;
	  dirs_stack = xrealloc (dirs_stack, sizeof (char *) * dirs_stack_size);
	  dirs_stack[dirs_stack_size - 1] = strdup (item_path);
	  }
	*/
	
	if (S_ISDIR(st.st_mode) &&
	    !isexcluded (item_path) && islocal (st.st_dev))
	  {
	    walkdir (item_path, check);
	  }
      }
    }
    
    /*
    for (; dirs_stack_size > 0; --dirs_stack_size) {
      walkdir (dirs_stack [dirs_stack_size - 1], check);
      free (dirs_stack [dirs_stack_size - 1]);
    }
    free (dirs_stack);
    */
    closedir (dp);
  }
  
  return;
}

#ifdef DEBUG
void print_stats (void) {
  fprintf (stderr,
	   "Total time: %u\n"
	   "Total objs: %u\n", (unsigned int)(t_end - t_start), checked_objects);
  return;
}
#endif

int main (void) {
  char *path;

  setlocale (LC_ALL, "");  
  textdomain (GETTEXT_DOMAIN);
  bindtextdomain (GETTEXT_DOMAIN, GETTEXT_LOCALEDIR);
  init_env (vars);

  if (rpmcheck_enabled == NULL) {
    rpmcheck_enabled = "0";
  } else {
    if (strcmp (rpmcheck_enabled, "0")) {
      if (rpmReadConfigFiles ((const char *) NULL, (const char *) NULL) != 0) {
	fprintf (stderr, _("rpmReadConfigFiles: Failed :[\n"));
	test_exit (E_FAIL, _("RPMCHECK is enabled and RPM configuration loading failed.\n"));
      }

      RPMts = rpmtsCreate ();
    }
  }

  if (selinuxcheck_enabled == NULL)
    selinuxcheck_enabled = "0";
  do_selinux_check = option_enabled(selinuxcheck_enabled);
#ifndef TEST_SELINUX
  if (do_selinux_check)
    test_exit (E_FAIL, _("SELINUXCHECK is enabled but filesystem test was not compiled with SELinux support.\n"));
#else
  // test file labels only if SELinux is enabled
  do_selinux_check = do_selinux_check && (is_selinux_enabled() == 1);
#endif

  init_symwhite ();
  init_devs (MTAB_PATH);

#ifdef DEBUG
  setbuf (stderr, NULL);
  setbuf (stdout, NULL);
  print_env (vars);
#endif

#ifdef DEBUG
  t_start = time(NULL);
#endif

  if (dirs_exclude != NULL) {
    if (strlen (dirs_exclude) > 0) {
      path = strtok (dirs_exclude, "\n");
      do {
	path = ltrim (path);
	if (strlen (path) > 0) {
	  ++dirs_exclude_arr_size;
	  dirs_exclude_arr = xrealloc (dirs_exclude_arr, sizeof (char *) * dirs_exclude_arr_size);
	  
	  { size_t len = strlen (path);
	    if (path[len-1] == '/' && len > 1) {
	      path[len-1] = '\0';
	    }}
	  
	  dirs_exclude_arr[dirs_exclude_arr_size - 1] = strdup (path);
	}
      } while ((path = strtok (NULL, "\n")) != NULL);
    }
  }
  
  if (dirs_include != NULL) {
#ifdef TEST_SELINUX
    matchpathcon_init(NULL);
#endif
    if (strlen (dirs_include) > 0) {
      path = strtok (dirs_include, "\n");
      do {
	path = ltrim (path);
	if (strlen (path) > 0) {
	  walkdir (path, &check_object);
	}
      } while ((path = strtok (NULL, "\n")) != NULL);
    }
#ifdef TEST_SELINUX
    matchpathcon_fini();
#endif
    /* else? */
  } else {
    report (INFO, 0, _("Nothing to check, INCLUDE variable is empty.\n"));
  }
  
#ifdef DEBUG
  t_end = time (NULL);
#endif

#ifdef DEBUG
  print_stats ();
#endif

  free_symwhite ();
  free(DEVid);
  test_exit (E_OK, NULL);
  return(E_FATAL);
}
